part of plot_window;

class LinePlotWindow extends PlotWindow {

  List<Line> lines = new List<Line>();


  NumSetting get startOfLinearRegressionSetting => new NumSetting("set start of linear regression", startOfLinearRegression, (num n) { this.startOfLinearRegression = n; });
  NumSetting get endOfLinearRegressionSetting => new NumSetting("set end of linear regression", endOfLinearRegression, (num n) { this.endOfLinearRegression = n; });
  
  num _startOfLinearRegression;
  num get startOfLinearRegression {
    _startOfLinearRegression ??= lines.map((Line l)=>l.smoothedPointList.first.x).reduce(math.min);
    return _startOfLinearRegression;
  }
  
  void set startOfLinearRegression(num n) {
    _startOfLinearRegression = n;
  }

  num _endOfLinearRegression;
  num get endOfLinearRegression {
    _endOfLinearRegression ??= lines.map((Line l)=>l.smoothedPointList.last.x).reduce(math.max); // (num found, num current)=>math.max(found, current));
    return _endOfLinearRegression;
  }
  
  void set endOfLinearRegression(num n) {
    _endOfLinearRegression = n;
  }
  
  
  
  LinePlotWindow(CanvasElement canvas):super(canvas);
  
  Iterable<Performer> getPerformers()=>[startOfLinearRegressionSetting, endOfLinearRegressionSetting];
  
  void addLine(Line line) {
    if (line.pointList == null || line.validPointList.points.isEmpty || lines.contains(line)) return;
    line.color??=colorIterator.getNext();
    /*if (line.color==null) {
      line.color=colorIterator.getNext();
      print("color picked: ${line.color}");
    }*/
    lines.add(line);
  }
  
  void removeData() {
    this.lines.clear();
  }
  
  void removeLine(String s) {
    this.lines.remove(s);
  }
  
  Rectangle getInitialSize()=>lines.map((Line l)=>l.rectangle).reduce((value, element)=>value.boundingBox(element));
  
  void drawContent() {
    for (Line line in lines) {
      var points = line.smoothedPointList.points;
      points = filterToRectangle(points);
      if (points.isNotEmpty) {
        if (line.lineType.hasLines) {
          context.save();
          context.setStrokeColorRgb(line.color.red, line.color.green, line.color.blue);
          context.lineWidth = line.width;
          Point first = toCanvas(points[0]);
          context..beginPath()
            ..moveTo(first.x, first.y);
          points.skip(1).map(toCanvas).forEach((Point p)=>context.lineTo(p.x, p.y));
          context.stroke();
          context.restore();
        }
        if (line.lineType.hasPoints) {
          context.save();
          context.fillStyle = "${line.color}";
          points.map(toCanvas).forEach((Point p) {
            context..moveTo(p.x, p.y)
              ..beginPath()
            ..arc(p.x, p.y, line.pointRadius, 0, 2*math.PI)
              ..fill();
          });
          context.restore();
        }
      }
      if (line.showLinearRegression) {
        if (line.cropLinearRegression)
          line.smoothedPointList.setLinearRegressionRange(rectangle.left, rectangle.right);
        else
          line.smoothedPointList.setLinearRegressionRange(startOfLinearRegression, endOfLinearRegression);
        var m = line.smoothedPointList.m;
        var c = line.smoothedPointList.c;
        if (m==null || m.isNaN || c==null || c.isNaN ) { continue; }
        Point first = toCanvas(new Point(rectangle.left, m * rectangle.left + c));
        context.save();
        context.setStrokeColorRgb(line.color.red, line.color.green, line.color.blue);
        context.lineWidth = line.width;
        context.setLineDash([5]);
        context..beginPath()
          ..moveTo(first.x, first.y);
        Point last = toCanvas(new Point(rectangle.right, m * rectangle.right + c));
        context.lineTo(last.x, last.y);
        context.stroke();
        context.restore();
      }
    }
    blankOutBorders();
  }

  Iterable<LegendEntry> getLegendEntries() {
    List<LegendEntry> legendEntries = new List<LegendEntry>();
    for(Line line in lines) {
      List<String> sublist = new List<String>()..add(line.name);
      if(line.showLinearRegression)
        sublist.add(line.smoothedPointList.getLinearRegressionString(line.linearRegressionFormat));
      legendEntries.add(new LegendEntry(line.color, sublist, line));
    }
    return legendEntries;
  }

}